import os
import crayons
import subprocess


def add_signature(func):
    def wrapper(msg=''):
        if not msg:
            return func(msg)
        else:
            return func('[*] %s' % msg)

    return wrapper


@add_signature
def success(msg):
    print(crayons.green(msg, bold=True))


@add_signature
def info(msg):
    print(crayons.white(msg, bold=True))


@add_signature
def error(msg):
    print(crayons.red(msg, bold=True))


@add_signature
def debug(msg):
    print(crayons.blue(msg, bold=False))


def exec_command(
    cmd,
    tf_data_dir=None,
    pre_func=lambda: None,
    except_func=lambda: None,
    else_func=lambda: None,
    finally_func=lambda: None,
):
    pre_func()

    if tf_data_dir:
        os.environ['TF_DATA_DIR'] = tf_data_dir

    try:
        subprocess.check_call(cmd)
    except subprocess.CalledProcessError:
        except_func()
    else:
        else_func()
    finally:
        os.environ.pop('TF_DATA_DIR', None)
        finally_func()



SAMPLE_CONFIG = {
    'environments': [
        {
            'name': 'e1',
            'aws_profile': 'e1_profile',
            'states_bucket': 'e1_bucket'
        },
        {
            'name': 'e2',
            'aws_profile': 'e2',
            'default': True,
            'states_bucket': 'e2_bucket'
        }
    ],
    'projects': {
        'p1': [
            'c1',
            'c2',
            {'c3': {'validate': {'check-variables': False}}},
            {
                'my-generic-component': [
                    'mg1',
                    'mg2',
                    {'mg3': {'validate': {'check-variables': False}}},
                ]
            },
            {'other-generic-component': ['oc1', 'oc2']},
        ],
        'p2': [
            'c1',
            {
                'c2': {'validate': {'check-variables': False}},
                'c3': {'validate': {'check-variables': False}},
            },
        ]
    }
}



#!/usr/bin/env python

from __future__ import print_function

import click

from config import get_plugins_dir
from config import parse_project_config

from install import install_terraform_plugin
from install import manual_install
from install import discover_and_install

from lib import run

from options import config_file_option
from options import components_option
from options import component_types_option
from options import environment_option
from options import exclude_components_option
from options import validate_project
from options import verbose_option

from utils import info


@click.group()
def main():
    """Give more power to Terraform."""
    pass


@config_file_option()
@main.command('projects')
def projects(config):
    """Display available projects in your configuration."""
    projects = config['projects'].keys()

    info('Available projects:')
    for project in projects:
        info('- %s' % project)


@config_file_option()
@component_types_option()
@click.argument('project', type=str)
@main.command('components')
def components(project, component_type, config):
    """Display available components for a project in your configuration."""
    project = validate_project(project, config)
    p_components = parse_project_config(config, project)
    components = p_components.keys()

    if component_type:
        info('Available components for project "%s" of type(s) "%s":' % (project, ', '.join(component_type)))
    else:
        info('Available components for project "%s":' % project)

    for component in components:
        should_print = True

        if component_type:
            component_config = p_components[component]
            ct = component_config.get('component_type', component)

            if ct not in component_type:
                should_print = False

        if should_print:
            info('- %s' % component)


@verbose_option()
@exclude_components_option()
@components_option()
@config_file_option()
@click.option(
    'project',
    '-p',
    '--project',
    type=str,
    default='',
    help='Project to manage.',
)
@click.argument('packages', type=str, nargs=-1)
@main.command('install')
def install(packages, **kwargs):
    """Discover and downloads plugins from your components."""
    plugins_dir = get_plugins_dir()

    install_terraform_plugin(plugins_dir)

    if packages:
        manual_install(plugins_dir)
    else:
        discover_and_install(plugins_dir, **kwargs)


@click.argument('tf_args', nargs=-1, type=click.UNPROCESSED)
@verbose_option()
@environment_option()
@exclude_components_option()
@components_option()
@config_file_option()
@click.argument('project', type=str)
@main.command('plan', context_settings=dict(ignore_unknown_options=True))
def tf_plan(project, components, exclude_components, environment, config, verbose, tf_args):
    """Terraform plan command (https://www.terraform.io/docs/commands/plan.html)."""
    run(
        command='plan',
        project=project,
        components=components,
        exclude_components=exclude_components,
        environment=environment,
        config=config,
        tf_args=tf_args,
        verbose=verbose,
    )


@click.argument('tf_args', nargs=-1, type=click.UNPROCESSED)
@verbose_option()
@config_file_option()
@environment_option()
@exclude_components_option()
@components_option()
@click.argument('project', type=str)
@main.command('apply', context_settings=dict(ignore_unknown_options=True))
def tf_apply(project, components, exclude_components, environment, config, verbose, tf_args):
    """Terraform apply command (https://www.terraform.io/docs/commands/apply.html)."""
    run(
        command='apply',
        project=project,
        components=components,
        exclude_components=exclude_components,
        environment=environment,
        config=config,
        tf_args=tf_args,
        verbose=verbose,
    )


@click.argument('tf_args', nargs=-1, type=click.UNPROCESSED)
@verbose_option()
@config_file_option()
@environment_option()
@exclude_components_option()
@components_option()
@click.argument('project', type=str)
@main.command('destroy', context_settings=dict(ignore_unknown_options=True))
def tf_destroy(project, components, exclude_components, environment, config, verbose, tf_args):
    """Terraform destroy command (https://www.terraform.io/docs/commands/destroy.html)."""
    run(
        command='destroy',
        project=project,
        components=components,
        exclude_components=exclude_components,
        environment=environment,
        config=config,
        tf_args=tf_args,
        verbose=verbose,
    )


@click.argument('tf_args', nargs=-1, type=click.UNPROCESSED)
@verbose_option()
@config_file_option()
@environment_option()
@exclude_components_option()
@components_option()
@click.argument('project', type=str)
@main.command('refresh', context_settings=dict(ignore_unknown_options=True))
def tf_refresh(project, components, exclude_components, environment, config, verbose, tf_args):
    """Terraform refresh command (https://www.terraform.io/docs/commands/refresh.html)."""
    run(
        command='refresh',
        project=project,
        components=components,
        exclude_components=exclude_components,
        environment=environment,
        config=config,
        tf_args=tf_args,
        verbose=verbose,
    )


@click.argument('tf_args', nargs=-1, type=click.UNPROCESSED)
@verbose_option()
@config_file_option()
@environment_option()
@exclude_components_option()
@components_option()
@click.argument('project', type=str)
@main.command('validate', context_settings=dict(ignore_unknown_options=True))
def tf_validate(project, components, exclude_components, environment, config, verbose, tf_args):
    """Terraform validate command (https://www.terraform.io/docs/commands/validate.html)."""
    run(
        command='validate',
        project=project,
        components=components,
        exclude_components=exclude_components,
        environment=environment,
        config=config,
        tf_args=tf_args,
        verbose=verbose,
    )


@click.argument('tf_args', nargs=-1, type=click.UNPROCESSED)
@verbose_option()
@config_file_option()
@environment_option()
@click.argument('id', type=str)
@click.argument('address', type=str)
@click.argument('component', type=str)
@click.argument('project', type=str)
@main.command('import', context_settings=dict(ignore_unknown_options=True))
def tf_import(project, component, address, id, environment, config, verbose, tf_args):
    """Terraform import command (https://www.terraform.io/docs/import/index.html)."""
    run(
        command='import',
        project=project,
        components=[component],
        environment=environment,
        config=config,
        tf_args=tf_args,
        verbose=verbose,
        address=address,
        id=id,
    )


@click.argument('tf_args', nargs=-1, type=click.UNPROCESSED)
@verbose_option()
@config_file_option()
@environment_option()
@exclude_components_option()
@components_option()
@click.argument('project', type=str)
@click.argument('subcommand', type=click.Choice(['list', 'mv', 'pull', 'push', 'rm', 'show']))
@main.command('state', context_settings=dict(ignore_unknown_options=True))
def tf_state(subcommand, project, components, exclude_components, environment, config, verbose, tf_args):
    """Terraform state command (https://www.terraform.io/docs/commands/state/index.html)."""
    run(
        command='state %s' % subcommand,
        project=project,
        components=components,
        exclude_components=exclude_components,
        environment=environment,
        config=config,
        tf_args=tf_args,
        verbose=verbose,
    )


if __name__ == '__main__':
    main()



import pytest

from stone_burner.config import parse_project_config
from stone_burner.config import TFAttributes

from utils import SAMPLE_CONFIG


def test_parse_project_config_1():
    e = {
        'c1': {'component_type': 'c1', 'validate': {}},
        'c2': {'component_type': 'c2', 'validate': {}},
        'c3': {'component_type': 'c3', 'validate': {'check-variables': False}},
        'mg1': {'component_type': 'my-generic-component', 'validate': {}},
        'mg2': {'component_type': 'my-generic-component', 'validate': {}},
        'mg3': {'component_type': 'my-generic-component', 'validate': {'check-variables': False}},
        'oc1': {'component_type': 'other-generic-component', 'validate': {}},
        'oc2': {'component_type': 'other-generic-component', 'validate': {}},
    }

    r = parse_project_config(SAMPLE_CONFIG, 'p1')

    assert r == e


def test_parse_project_config_2():
    with pytest.raises(Exception):
        parse_project_config(SAMPLE_CONFIG, 'p2')

    def check_variables(*args, **kwargs):
        component_config = kwargs['component_config']
        validate_config = component_config.get('validate', {})
        check_variables = validate_config.get('check-variables', True)

        return ['true'] if check_variables else ['false']


def test_TFAttributes_check_variables_1():
    tf_attrs = TFAttributes()
    r = tf_attrs.check_variables(
        component_config={'component_type': 'c3', 'validate': {'check-variables': True}})

    assert r == ['true']


def test_TFAttributes_check_variables_2():
    tf_attrs = TFAttributes()
    r = tf_attrs.check_variables(
        component_config={'component_type': 'c3', 'validate': {'check-variables': False}})

    assert r == ['false']


def test_TFAttributes_check_variables_3():
    tf_attrs = TFAttributes()
    r1 = tf_attrs.check_variables(
        component_config={'component_type': 'c3'})

    assert r1 == ['true']

    r2 = tf_attrs.check_variables(
        component_config={'component_type': 'c3', 'validate': {}})

    assert r2 == ['true']

    r3 = tf_attrs.check_variables(
        component_config={'component_type': 'c3', 'validate': None})

    assert r3 == ['true']



import click
import os
import yaml

from config import parse_project_config


def config_file_option():
    class Config(click.ParamType):
        name = 'config'

        def convert(self, value, param, ctx):
            if not os.path.exists(value):
                self.fail('%s does not exist' % value)

            with open(value) as f:
                config = yaml.load(f)

            ctx.params['config'] = config
            return config

    return click.option(
        '-config',
        '--config-file',
        'config',
        type=Config(),
        default='config.yml',
        help='Configuration file to use.',
        show_default=True,
    )


def environment_option():
    return click.option(
        '-e',
        '--environment',
        type=str,
        help='Target environment (AWS account to use). '
             'Remember that your environment must be defined in your config.yml file.',
    )


def components_option():
    return click.option(
        '-c',
        '--components',
        type=str,
        default=[],
        help='Individual components to manage.',
        multiple=True,
    )


def exclude_components_option():
    return click.option(
        '-xc',
        '--exclude-components',
        type=str,
        default=[],
        help='Individual components to exclude.',
        multiple=True,
    )


def component_types_option():
    return click.option(
        '-ct',
        '--component-type',
        type=str,
        default=[],
        help='Apply your action only to the components which belong to this type.',
        multiple=True,
    )


def verbose_option():
    return click.option('-v', '--verbose', count=True)


def validate_project(project, config):
    valid_projects = config['projects'].keys()

    if project not in valid_projects:
        err_msg = 'invalid choice: %s. (choose from %s)' % (
            project,
            ', '.join(valid_projects),
        )

        raise click.BadParameter(err_msg, param_hint='"project"')

    return project


def validate_components(components, project, config):
    valid_components = parse_project_config(config, project).keys()

    # Avoid duplicates
    components = list(set(components))

    for component in components:
        if component not in valid_components:
            err_msg = 'invalid choice: %s. (choose from %s)' % (
                component,
                ', '.join(valid_components),
            )

            raise click.BadParameter(err_msg, param_hint='"components"')

    return components


def validate_environment(environment, config):
    env_config = config['environments']

    if not environment:
        environment = [env['name'] for env in env_config if env.get('default', False)][0]

    valid_envs = [env['name'] for env in env_config]

    if environment not in valid_envs:
        err_msg = 'invalid choice: %s. (choose from %s)' % (
            environment,
            ', '.join(valid_envs),
        )

        raise click.BadParameter(err_msg, param_hint='"environment"')

    return environment




import pytest
import click

from stone_burner.options import validate_components
from stone_burner.options import validate_environment

from utils import SAMPLE_CONFIG


def test_validate_components_1():
    e = ['c1', 'c2', 'oc1']
    r = validate_components(['c1', 'c2', 'oc1'], 'p1', SAMPLE_CONFIG)

    assert len(r) == len(e) and sorted(r) == sorted(e)


def test_validate_components_2():
    with pytest.raises(click.BadParameter):
        validate_components(['c1', 'c2', 'oc1', 'ne-c'], 'p1', SAMPLE_CONFIG)


def test_validate_environment_1():
    assert 'e1' == validate_environment('e1', SAMPLE_CONFIG)


def test_validate_environment_2():
    with pytest.raises(click.BadParameter):
        validate_environment('e3', SAMPLE_CONFIG)



import os
import subprocess

from stone_burner.lib import run
from stone_burner.lib import check_validation

from utils import SAMPLE_CONFIG


def test_run_1(monkeypatch):
    """Test run does not crash."""
    monkeypatch.setattr(subprocess, 'check_call', lambda cmd: True)
    monkeypatch.setattr(os, 'chdir', lambda path: True)

    run(
        command='plan',
        project='p1',
        components=['c1', 'c2', 'c3'],
        exclude_components=[],
        environment='e1',
        config=SAMPLE_CONFIG,
        tf_args=[],
        verbose=0,
    )


def test_run_2(monkeypatch):
    """Test run does not crash."""
    monkeypatch.setattr(subprocess, 'check_call', lambda cmd: True)
    monkeypatch.setattr(os, 'chdir', lambda path: True)

    run(
        command='validate',
        project='p1',
        components=['c1', 'c2', 'c3'],
        exclude_components=[],
        environment='e1',
        config=SAMPLE_CONFIG,
        tf_args=[],
        verbose=0,
    )


def test_check_validation_1(monkeypatch):
    monkeypatch.setattr(os.path, 'exists', lambda path: True)
    r = check_validation('p1', 'c1', 'e1', {'component_type': 'ct1', 'validate': {}})

    assert r is True


def test_check_validation_2(monkeypatch):
    monkeypatch.setattr(os.path, 'exists', lambda path: True)
    r = check_validation('p1', 'c1', 'e1', {'component_type': 'ct1', 'validate': {'skip': True}})

    assert r is False


def test_check_validation_3(monkeypatch):
    monkeypatch.setattr(os.path, 'exists', lambda path: False)
    r = check_validation('p1', 'c1', 'e1', {'component_type': 'ct1', 'validate': {}})

    assert r is False



import os
import six


DEFAULT_STONE_BURNER_DIR = os.path.join(os.path.expanduser("~"), '.stoneburner')
DEFAULT_ROOT_DIR = os.path.abspath(os.getcwd())
DEFAULT_STATES_DIR = os.path.abspath('./states')
DEFAULT_PROJECTS_DIR = os.path.abspath('./projects')
DEFAULT_VARS_DIR = os.path.abspath('./variables')

OPTIONS_BY_COMMAND = {
    'init': {
        'options': ['backend', 'backend-config', 'plugin-dir', 'get-plugins'],
        'args': []
    },
    'get': {
        'options': [],
        'args': []
    },
    'plan': {
        'options': ['var-file', 'state'],
        'args': [],
    },
    'apply': {
        'options': ['var-file', 'state'],
        'args': [],
    },
    'destroy': {
        'options': ['var-file', 'state'],
        'args': [],
    },
    'refresh': {
        'options': ['var-file', 'state'],
        'args': [],
    },
    'import': {
        'options': ['var-file', 'state'],
        'args': ['address', 'id'],
    },
    'validate': {
        'options': ['var-file', 'check-variables'],
        'args': [],
    },
    'state': {
        'options': [],
        'args': [],
    },
}


def parse_project_config(config, project):
    result = {}
    p_config = config['projects'][project]

    for elem in p_config:
        if isinstance(elem, six.string_types):
            result[elem] = {
                'component_type': elem,
                'validate': {},
            }
        elif isinstance(elem, dict):
            d_keys = elem.keys()

            if len(d_keys) != 1:
                raise Exception(
                    'Bad config for project: components defined as dictionaries must have exactly 1 key')

            component_type = d_keys[0]  # The only key is the component type
            info = elem[component_type]

            if isinstance(info, list):
                # It's a list of components from the same type
                for c in info:
                    if isinstance(c, six.string_types):
                        result[c] = {
                            'component_type': component_type,
                            'validate': {},
                        }
                    elif isinstance(c, dict):
                        d_keys = c.keys()

                        if len(d_keys) != 1:
                            raise Exception(
                                'Bad config for project: components defined as dictionaries must have exactly 1 key')

                        # The only key is the component type
                        component_name = d_keys[0]
                        c_info = c[component_name]

                        result[component_name] = {
                            'component_type': component_type,
                            'validate': c_info['validate'],
                        }
            elif isinstance(info, dict):
                # It's a component named the same as the component type with extra info
                result[component_type] = {
                    'component_type': component_type,
                    'validate': info['validate'],
                }

    return result


def get_plugins_dir():
    stone_burner_dir = os.environ.get(
        'STONE_BURNER_DIR', DEFAULT_STONE_BURNER_DIR
    )

    plugin_dir = os.path.join(stone_burner_dir, 'plugins')

    if not os.path.exists(stone_burner_dir):
        os.makedirs(stone_burner_dir)

    if not os.path.exists(plugin_dir):
        os.makedirs(plugin_dir)

    return plugin_dir


def get_component_paths(project, component, component_config, environment, states_dir=DEFAULT_STATES_DIR, projects_dir=DEFAULT_PROJECTS_DIR, vars_dir=DEFAULT_VARS_DIR):
    return {
        'state_dir': os.path.abspath(os.path.join(states_dir, environment, project, component)),
        'config_dir': os.path.abspath(os.path.join(projects_dir, project, component_config['component_type'])),
        'vars_file': os.path.abspath(os.path.join(vars_dir, environment, project, '%s.tfvars' % component)),
    }


class TFAttributes(object):
    def __init__(
        self,
        root_dir=DEFAULT_ROOT_DIR,
        projects_dir=DEFAULT_PROJECTS_DIR,
        states_dir=DEFAULT_STATES_DIR,
        vars_dir=DEFAULT_VARS_DIR
    ):
        self.root_dir = root_dir
        self.projects_dir = projects_dir
        self.states_dir = states_dir
        self.vars_dir = vars_dir
        self.plugin_dir = get_plugins_dir()

    def backend(*args, **kwargs):
        no_remote = os.environ.get('TF_NO_REMOTE', '0')
        return ['false'] if no_remote == '1' else ['true']

    def backend_config(*args, **kwargs):
        project = kwargs['project']
        component = kwargs['component']
        environment = kwargs['environment']
        config = kwargs['config']

        state_key = os.path.join(environment, project, '%s.tfstate' % component)

        env_config = {
            env['name']: [
                'bucket=%s' % env['states_bucket'],
                'profile=%s' % env['aws_profile'],
                'key=%s' % state_key,
            ]
            for env in config['environments']
        }

        return env_config[environment]

    def plugin_dir(self, *args, **kwargs):
        return [self.plugin_dir]

    def get_plugins(*args, **kwargs):
        return ['false']

    def var_file(self, *args, **kwargs):
        project = kwargs['project']
        component = kwargs['component']
        environment = kwargs['environment']
        component_config = kwargs['component_config']

        result = []

        shared_vars_file = os.path.join(self.vars_dir, environment, project, 'shared.tfvars')
        variables = component_config.get('variables', component)
        vars_file = os.path.join(self.vars_dir, environment, project, '%s.tfvars' % variables)

        if os.path.exists(shared_vars_file):
            result.append(shared_vars_file)

        if os.path.exists(vars_file):
            result.append(vars_file)

        return result

    def address(*args, **kwargs):
        return [kwargs['address']]

    def id(*args, **kwargs):
        return [kwargs['id']]

    def check_variables(*args, **kwargs):
        component_config = kwargs['component_config']
        validate_config = component_config.get('validate', {}) or {}
        check_variables = validate_config.get('check-variables', True)

        return ['true'] if check_variables else ['false']

    def state(self, *args, **kwargs):
        project = kwargs['project']
        component = kwargs['component']
        environment = kwargs['environment']

        state_file = os.path.join(
            self.states_dir, environment, project, component, 'terraform.tfstate'
        )

        return [state_file]



# -*- coding: utf-8 -*-

from setuptools import find_packages
from setuptools import setup

setup(
    name='stone-burner',
    version='0.7.1',
    description='Give more power to Terraform.',
    author='Rodrigo Argüello Flores',
    author_email='rodrigo@kkvesper.jp',
    url='https://github.com/kkvesper/stone-burner',
    download_url='https://github.com/kkvesper/stone-burner/archive/v0.7.1.tar.gz',
    packages=find_packages(),
    install_requires=[
        'jinja2>=2.9',
        'pyyaml>=3.12',
        'click>=6.7',
        'crayons>=0.1.2',
    ],
    entry_points={
        'console_scripts': [
            'stone-burner = stone_burner.cli:main',
        ],
    },
    keywords=['terraform', 'planetary-annihilation', 'infrastructure-as-code'],
)



from __future__ import print_function

import copy
import os
import sys

from config import OPTIONS_BY_COMMAND
from config import TFAttributes
from config import parse_project_config
from config import get_component_paths

from options import validate_components
from options import validate_environment
from options import validate_project

from utils import error
from utils import exec_command
from utils import info
from utils import success


def run_command(cmd, project, component, component_config, environment, verbose=0, *args, **kwargs):
    work_dir = os.getcwd()
    c_paths = get_component_paths(project, component, component_config, environment)
    state_dir = c_paths['state_dir']
    config_dir = c_paths['config_dir']

    new_kwargs = copy.deepcopy(kwargs)
    new_kwargs['tf_args'] = []  # Don't want to send extra params to get and init commands

    need_init = (
        os.environ.get('TF_INIT', '0') == '1' or
        not os.path.exists(state_dir) or
        'terraform.tfstate' not in os.listdir(state_dir) or
        'plugins' not in os.listdir(state_dir) or
        'plugin_path' not in os.listdir(state_dir)
    )

    os.chdir(config_dir)

    def handle_init_error():
        error('There was an error executing your command. Please check the Terraform output.')
        sys.exit(1)

    def pre_cmd_msg():
        if verbose >= 0:
            info('Running Terraform command: %s' % ' '.join(cmd))
            info('<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>')
            info('           COMMAND OUTPUT')
            info('')

    def handle_cmd_success():
        if verbose >= 0:
            success('OK!')

    def handle_cmd_error():
        error('')
        error('There was an error executing your command. Please check the Terraform output.')
        sys.exit(1)

    def handle_cmd_end():
        if verbose >= 0:
            info('<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>')

    if need_init:
        init_tf_cmd = 'init'

        def init_pre_func():
            info('State not found or init forced, Initializing with terraform init...') if verbose >= 0 else None
    else:
        init_tf_cmd = 'get'

        def init_pre_func():
            info('Fetching modules with terraform get...') if verbose >= 0 else None

    init_cmd = build_command(
        *args,
        command=init_tf_cmd,
        project=project,
        component=component,
        component_config=component_config,
        environment=environment,
        **new_kwargs
    )

    exec_command(
        cmd=init_cmd,
        tf_data_dir=state_dir,
        pre_func=init_pre_func,
        except_func=handle_init_error,
        else_func=handle_cmd_success,
    )

    exec_command(
        cmd=cmd,
        tf_data_dir=state_dir,
        pre_func=pre_cmd_msg,
        except_func=handle_cmd_error,
        else_func=handle_cmd_success,
        finally_func=handle_cmd_end,
    )

    os.chdir(work_dir)


def build_command(command, tf_args=[], options_by_command=OPTIONS_BY_COMMAND, *args, **kwargs):
    options = []
    arguments = []

    if ' ' in command:
        commands = command.split()
        command = commands[0]
        subcommands = commands[1:]
    else:
        subcommands = []

    for option in options_by_command.get(command, {}).get('options', []):
        func_name = option.replace('-', '_')
        func = getattr(TFAttributes, func_name)

        values = func(TFAttributes(), *args, **kwargs)
        options += ['-%s=%s' % (option, value) for value in values]

    for arg in options_by_command.get(command, {}).get('args', []):
        func_name = arg.replace('-', '_')
        func = getattr(TFAttributes, func_name)

        values = func(TFAttributes(), *args, **kwargs)
        arguments += values

    return ['terraform', command] + subcommands + options + list(tf_args) + arguments


def check_validation(project, component, environment, component_config, verbose=0):
    title = '%s %s - %s %s' % ('=' * 10, project, component, '=' * 10)
    c_paths = get_component_paths(project, component, component_config, environment)

    vars_file = c_paths['vars_file']

    if verbose >= 0:
        info(title)

    # Don't validate projects without variables file
    if not os.path.exists(vars_file):
        if verbose >= 0:
            info('Skipping validation. Reason: vars file "%s" not found.' % vars_file)
            success('OK!')

        return False

    validate_config = component_config.get('validate', None)

    if validate_config and 'skip' in validate_config:
        if verbose >= 0:
            info('Skipping validation. Reason: "skip" found in the configuration.')
            success('OK!')

        return False

    return True


def run(command, project, components, environment, config, exclude_components=[], verbose=0, *args, **kwargs):
    project = validate_project(project, config)
    p_components = parse_project_config(config, project)

    if components:
        components = validate_components(components, project, config)
    else:
        # If no component is chosen, use all of them
        components = p_components.keys()

    components = list(set(components) - set(exclude_components))
    environment = validate_environment(environment, config)

    for component in components:
        component_config = p_components[component] or {}

        if command == 'validate':
            should_validate = check_validation(
                project=project,
                component=component,
                environment=environment,
                component_config=component_config,
                verbose=verbose,
            )

            if not should_validate:
                continue

        cmd = build_command(
            *args,
            command=command,
            project=project,
            component=component,
            component_config=component_config,
            environment=environment,
            config=config,
            verbose=verbose,
            **kwargs
        )

        if verbose >= 0:
            info('::::::::::::::::::::::::::::::::::::::::::::::')
            info('PROJECT =======> %s - %s' % (project, component))

        run_command(
            *args,
            cmd=cmd,
            project=project,
            component=component,
            component_config=component_config,
            environment=environment,
            config=config,
            verbose=verbose,
            **kwargs
        )

        if verbose >= 0:
            info('::::::::::::::::::::::::::::::::::::::::::::::')
            info('')
            info('')



import os
import stat
import subprocess
import shutil
import tempfile
import platform
import urllib
import zipfile

from config import parse_project_config
from config import get_component_paths

from options import validate_project
from options import validate_components

from utils import info
from utils import success
from utils import error
from utils import exec_command

# 0755
PLUGIN_PERMISSIONS = (
    stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR |
    stat.S_IRGRP | stat.S_IXGRP |
    stat.S_IROTH | stat.S_IXOTH
)


def install_terraform_plugin(plugin_dir, plugin_permissions=PLUGIN_PERMISSIONS):
    info('Installing terraform provider plugin from terraform binary...')

    for plugin in os.listdir(plugin_dir):
        if plugin.startswith('terraform-provider-terraform'):
            # Remove previous terraform plugin
            os.remove(os.path.join(plugin_dir, plugin))

    tf_bin = subprocess.check_output(['which', 'terraform']).split('\n')[0]
    tf_version = subprocess.check_output(
        ['terraform', '-v']).split('\n')[0].split(' ')[1]

    info('Found terraform at %s on version: %s' % (tf_bin, tf_version))

    tf_plugin_name = 'terraform-provider-terraform_%s_x4' % tf_version
    tf_plugin_path = os.path.join(plugin_dir, tf_plugin_name)

    info('Installing %s on %s' % (tf_plugin_name, plugin_dir))
    shutil.copy2(tf_bin, tf_plugin_path)
    os.chmod(tf_plugin_path, plugin_permissions)
    success('OK!')


def manual_install(packages, plugin_dir, plugin_permissions=PLUGIN_PERMISSIONS):
    suffix = ''
    system = platform.system()

    if system == 'Darwin':
        suffix = 'darwin_amd64'
    elif system == 'Linux':
        machine = platform.machine()

        if machine == 'x86_64':
            suffix = 'linux_amd64'
        elif machine == 'i386':
            suffix = 'linux_386'
        else:
            raise Exception('Unsupported Linux architecture: %s' % machine)
    else:
        raise Exception('Unsupported distribution: %s' % system)

    base_url = 'https://releases.hashicorp.com/'

    temp_dir = tempfile.mkdtemp()
    downloader = urllib.URLopener()

    for pkg in packages:
        info('Installing %s...' % pkg)
        try:
            name, version = pkg.split('@')
        except ValueError:
            error('Bad syntax: %s.' % pkg)
            error(
                'Packages must be specified with the following syntax: <name>@<version>')
        else:
            fname = 'terraform-provider-%s_%s_%s.zip' % (
                name, version, suffix)
            url = os.path.join(
                base_url, 'terraform-provider-%s' % name, version, fname)
            dest_file = os.path.join(temp_dir, fname)

            info('downloading %s...' % url)
            try:
                downloader.retrieve(url, dest_file)
            except Exception:
                import traceback
                error('An error ocurred downloading %s' % url)
                error(traceback.format_exc())
            else:
                info('Extracting %s to %s...' % (fname, plugin_dir))
                zip_ref = zipfile.ZipFile(dest_file, 'r')
                zip_ref.extractall(plugin_dir)
                zip_ref.close()
                success('OK!')

    shutil.rmtree(temp_dir)
    info('Setting plugin permissions...')
    for f in os.listdir(plugin_dir):
        os.chmod(os.path.join(plugin_dir, f), plugin_permissions)

    success('OK!')


def discover_and_install(plugins_dir, project, components, exclude_components, config, verbose):
    project = validate_project(project, config)
    p_components = parse_project_config(config, project)

    if components:
        components = validate_components(components, project, config)
    else:
        # If no component is chosen, use all of them
        components = p_components.keys()

    components = list(set(components) - set(exclude_components))

    workdir = os.getcwd()

    for component in components:
        component_config = p_components[component] or {}
        c_paths = get_component_paths(project, component, component_config, environment)
        config_dir = c_paths['config_dir']

        os.chdir(config_dir)

        temp_dir = tempfile.mkdtemp()

        exec_command(
            cmd=['terraform', 'init', '-backend=false', '-get=true', '-get-plugins=true', '-input=false'],
            tf_data_dir=temp_dir,
        )

        for root, dirs, filenames in os.walk(os.path.join(temp_dir, 'plugins')):
            for f in filenames:
                f_path = os.path.join(root, f)

                if f != 'lock.json':
                    # TODO: keep json.lock file and merge new ones
                    shutil.move(f_path, os.path.join(plugins_dir, f))

        shutil.rmtree(temp_dir)
        os.chdir(workdir)
